package nl.tue.astar.impl.cached;

import java.io.IOException;

import nl.tue.astar.AStarException;
import nl.tue.astar.AStarObserver;
import nl.tue.astar.Head;
import nl.tue.astar.Record;
import nl.tue.astar.Trace;
import nl.tue.astar.impl.State;
import nl.tue.astar.impl.memefficient.MemoryEfficientAStarAlgorithm;
import nl.tue.astar.impl.memefficient.MemoryEfficientAStarThread;
import nl.tue.storage.impl.SkippableOutputStream;

public class CachedMemoryEfficientAStarThread<H extends Head, T extends FastLowerBoundTail>
		extends MemoryEfficientAStarThread<H, T> {

	public CachedMemoryEfficientAStarThread(
			MemoryEfficientAStarAlgorithm<H, T> algorithm, H initialHead,
			Trace trace, int maxStates) {
		super(algorithm, initialHead, trace, maxStates);
	}

	@Override
	protected Record poll() {
		Record node = queue.poll();
		return node;
	}

	@Override
	protected void processMovesForRecord(Record rec, H head, T tail,
			int stopAt, double timeLimit, long endTime) throws AStarException {
		int storedEstimate = rec.getEstimatedRemainingCost();
		if (storedEstimate < 0) {
			// must be the initial state.
			// rec.setEstimatedRemainingCost(tailEstimate);

			considered.add(rec.getState());
			for (AStarObserver observer : observers) {
				observer.nodeVisited(rec);
			}
			super.processMovesForRecord(rec, head, tail, stopAt, timeLimit,
					endTime);

		} else {
			boolean wasExact = tail.isExactEstimateKnown();

			if (!wasExact) {
				reComputeCachedTail(rec, tail, head, storedEstimate);
			}
			int tailEstimate = tail.getEstimatedCosts(delegate, head);

			if (storedEstimate < tailEstimate) {
				// the tail's true estimate is higher than the fast one stored
				// earlier. Re-queue the record
				rec.setEstimatedRemainingCost(tailEstimate);

				// reduce the pollCount
				poll--;

				// Do not count the queued state, since it is a re-queuing.
				if (queue.add(rec)) {
					queuedStateCount++;
				}

				// edge from oldRec to newRec traversed.
				for (AStarObserver observer : observers) {
					observer.edgeTraversed(rec, rec);
				}
			} else {
				considered.add(rec.getState());
				for (AStarObserver observer : observers) {
					observer.nodeVisited(rec);
				}
				super.processMovesForRecord(rec, head, tail, stopAt, timeLimit,
						endTime);
			}
		}
	}

	protected void reComputeCachedTail(Record rec, T tail, H head,
			int storedEstimate) throws AStarException {
		for (AStarObserver observer : observers) {
			observer.estimateComputed(head);
		}
		computedEstimateCount++;

		tail.computeEstimate(delegate, head, storedEstimate);
		// The tail object may have changed. Update in storage!
		assert tail.isExactEstimateKnown();
		synchronized (statespace) {
			SkippableOutputStream out = store.getOutputStreamForObject(rec
					.getState());
			// rewrite the tail in this outputstream
			try {
				((CachedStorageAwareDelegate<H, T>) delegate).getHeadDeflater()
						.skip(head, out);
				((CachedStorageAwareDelegate<H, T>) delegate).getTailDeflater()
						.deflate(tail, out);
			} catch (IOException e) {
				throw new AStarException(e);
			}
		}
	}

	@Override
	@SuppressWarnings("unchecked")
	protected T computeNewTail(T tail, H newHead, int modelMove,
			int movedEvent, int activity) {
		T newTail = (T) tail.getNextTail(delegate, newHead, modelMove,
				movedEvent, activity);

		if (newTail.isExactEstimateKnown()) {
			for (AStarObserver observer : observers) {
				observer.estimateComputed(newHead);
			}
			computedEstimateCount++;
		}

		return newTail;
	}

	protected void storeStateForRecord(State<H, T> state, Record newRec)
			throws AStarException {
		super.storeStateForRecord(state, newRec);
	}

}
